"""
Holds general information about the project.
"""

__title__ = 'cogpyt'
__author__ = 'Boyan Hristov'
__description__ = "Code Generation in (pure) Python"
__url__ = 'https://github.com/BoyanH/cogpyt'
__version__ = '0.0.1'
__copyright__ = 'Copyright 2022 Boyan Hristov'
__licence__ = 'MIT'
